import numpyro.distributions as dist
import jax.numpy as jnp
from jax import random, vmap
from jax.lax import scan

from lqg.riccati import kalman_gain
from lqg.model import Dynamics


class KalmanFilter:
    def __init__(self, dynamics: Dynamics):
        self.dynamics = dynamics

    @property
    def xdim(self):
        return self.dynamics.A.shape[0] * 2

    def K(self, T):
        K = kalman_gain(self.dynamics.A, self.dynamics.C,
                        self.dynamics.V @ self.dynamics.V.T,
                        self.dynamics.W @ self.dynamics.W.T, T=T)
        return K

    def simulate(self, rng_key, n=1, T=100, x0=None, xhat0=None):
        """ Simulate n trials

        Args:
            rng_key (jax.random.PRNGKey): random number generator key
            n (int): number of trials
            T (int): number of time steps
            x0 (jnp.array): initial state
            xhat0 (jnp.array): initial belief

        Returns:
            jnp.array (T, n, d)
        """

        # compute Kalman gain
        K = self.K(T)

        def simulate_trial(rng_key, T=100, x0=None, xhat0=None):
            """ Simulate a single trial

            Args:
                rng_key (jax.random.PRNGKey): random number generator key
                T (int): number of time steps
                x0 (jnp.array): initial state
                xhat0 (jnp.array): initial belief

            Returns:
                jnp.array, jnp.array, jnp.array, jnp.array: x (states), x_hat (estimates), y, u
            """

            x0 = jnp.zeros(self.dynamics.A.shape[0]) if x0 is None else x0
            xhat0 = jnp.zeros(self.dynamics.A.shape[0]) if xhat0 is None else xhat0

            # generate standard normal noise terms
            rng_key, subkey = random.split(rng_key)
            epsilon = random.normal(subkey, shape=(T, x0.shape[0]))
            rng_key, subkey = random.split(rng_key)
            eta = random.normal(subkey, shape=(T, x0.shape[0]))

            def loop(carry, t):
                x, x_hat = carry

                # apply dynamics
                x = self.dynamics.A @ x + self.dynamics.V @ epsilon[t]

                # generate observation
                y = self.dynamics.C @ x + self.dynamics.W @ eta[t]

                # update agent's belief
                x_pred = self.dynamics.A @ x_hat
                x_hat = x_pred + K @ (y - self.dynamics.C @ x_pred)

                return (x, x_hat), (x, x_hat, y)

            _, (x, x_hat, y) = scan(loop, (x0, xhat0), jnp.arange(1, T))

            x = jnp.vstack((x0, x))
            x_hat = jnp.vstack((xhat0, x_hat))

            # stack up state and estimate
            # so that the data have the same structure as the control models
            x = jnp.hstack((x, x_hat))

            # rearrange dimension so that we always have the state
            # and then the corresponding estimate (interleaved)
            return jnp.hstack([x[..., ::2], x[..., 1::2]])

        # simulate n trials
        x = vmap(lambda key: simulate_trial(key, T=T, x0=x0, xhat0=xhat0),
                 out_axes=1)(random.split(rng_key, num=n))

        return x

    def conditional_distribution(self, x):
        T, n, d = x.shape
        idx = list(range(d))
        swapdim = idx[::2] + idx[1::2]

        V = self.dynamics.V
        W = self.dynamics.W

        K = self.K(T)

        F = jnp.vstack(
            [jnp.hstack([self.dynamics.A, jnp.zeros_like(self.dynamics.A)]),
             jnp.hstack([K @ self.dynamics.C @ self.dynamics.A,
                         self.dynamics.A - K @ self.dynamics.C @ self.dynamics.A])])

        F = F[:, swapdim]
        F = F[swapdim, :]

        # TODO: this throws errors for the model with velocities..
        G = jnp.vstack([jnp.hstack([V, jnp.zeros_like(V)]), jnp.hstack([K @ self.dynamics.C @ V, K @ W])])

        G = G[:, swapdim]
        G = G[swapdim, :]

        mu = jnp.zeros((n, d))
        Sigma = G @ G.T

        def f(carry, xt):
            mu, Sigma = carry

            mu = mu @ F.T + (xt - mu) @ jnp.linalg.inv(Sigma).T @ (F @ Sigma).T

            Sigma = F @ Sigma @ F.T + G @ G.T - (F @ Sigma) @ jnp.linalg.inv(Sigma) @ (Sigma @ F.T)
            return (mu, Sigma), (mu, Sigma)

        _, (mu, Sigma) = scan(f, (mu, Sigma), x)

        return dist.MultivariateNormal(mu.transpose((1, 0, 2)), Sigma)

    def log_likelihood(self, x):
        return self.conditional_distribution(x[:-1]).log_prob(x[1:].transpose((1, 0, 2)))



import jax.numpy as jnp
from jax.lax import scan


def solve_discrete_riccati(A, B, Q, R, T):
    """ Solve a discrete-time algebraic Riccati equation (DARE)
    by iterating

    S = A' (S - S B (B' S B + R)^-1 B' S) A + Q

    Args:
        A (jnp.array): square matrix
        B (jnp.array): matrix
        Q (jnp.array): matrix
        R (jnp.array): square matrix
        T (jnp.array): number of time steps

    Returns:
        jnp.array: solution of the DARE
    """

    def riccati_iter(S, t):
        S = A.T @ (S - S @ B @ jnp.linalg.inv(B.T @ S @ B + R) @ B.T @ S) @ A + Q
        return S, S

    S, _ = scan(riccati_iter, Q, jnp.arange(T))

    return S


def kalman_gain(A, C, V, W, T):
    """ Compute Kalman gain after T time steps

    Args:
        A (jnp.array): state transition matrix
        C (jnp.array): observation matrix
        V (jnp.array): state transition noise covariance
        W (jnp.array): observation noise covariance
        T (int): number of time steps

    Returns:
        jnp.array: Kalman gain (final time step)
    """
    P = solve_discrete_riccati(A.T, C.T, V, W, T)

    S = C @ P @ C.T + W
    K = P @ C.T @ jnp.linalg.inv(S)

    return K


def control_law(A, B, Q, R, T):
    """ LQR control law after T time steps

    Args:
        A (jnp.array): state transition matrix
        B (jnp.array): control matrix
        Q (jnp.array): control costs
        R (jnp.array): action costs
        T (jnp.array): time steps

    Returns:
        jnp.array: LQR controller
    """
    S = solve_discrete_riccati(A, B, Q, R, T)

    L = jnp.linalg.inv(B.T @ S @ B + R) @ B.T @ S @ A

    return L



from jax import numpy as jnp
from jax.scipy import linalg

from lqg.model import System, Actor, Dynamics


class SubjectiveModel(System):
    def __init__(self, process_noise=1., c=0.5, motor_noise=0.5, subj_noise=1.,
                 sigma=6., prop_noise=3., dt=1. / 60):
        A = jnp.eye(2)
        B = jnp.array([[0.], [10. * dt]])
        C = jnp.eye(2)

        V = jnp.diag(jnp.array([process_noise, motor_noise]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))
        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)

        A = jnp.eye(2)
        B = jnp.array([[0.], [10. * dt]])
        # C = jnp.array([[1., -1.]])
        C = jnp.eye(2)

        V = jnp.diag(jnp.array([subj_noise, motor_noise]))

        Q = jnp.array([[1., -1.], [-1., 1.]])
        R = jnp.eye(B.shape[1]) * c

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class SubjectiveVelocityModel(System):
    def __init__(self, process_noise=1., c=0.5, motor_noise=0.5, subj_noise=.1, subj_vel_noise=10.,
                 sigma=6., prop_noise=3., dt=1. / 60):
        A = jnp.eye(2)
        B = jnp.array([[0.], [10. * dt]])
        # A = linalg.block_diag(*[jnp.array([[1.]]), jnp.array([[1., dt], [0., 1.]])])
        # B = jnp.array([[0.], [0.], [10. * dt]])
        # C = jnp.array([[1., -1.]])
        C = jnp.eye(2)
        # C = jnp.array([[1., 0, 0.], [0., 1., 0.]])

        V = jnp.diag(jnp.array([process_noise, motor_noise]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)

        A = jnp.array([[1., 0., dt], [0., 1., 0.], [0., 0., 1.]])
        # A = jnp.array([[1., 0., dt, 0.], [0., 1., 0., dt], [0., 0., 1., 0.], [0., 0., 0., 1.]])
        B = jnp.array([[0.], [10. * dt], [0.]])
        # B = jnp.array([[0.], [0.], [0.], [10. * dt]])
        C = jnp.array([[1., 0., 0.],
                       [0., 1., 0.]])
        # C = jnp.array([[1., 0., 0., 0.],
        #                [0., 1., 0., 0.]])
        # C = jnp.array([[1., -1., 0.]])

        V = jnp.diag(jnp.array([subj_noise, motor_noise, subj_vel_noise]))

        # V = linalg.cholesky(jnp.array([[dt ** 2 / 3 * subj_noise ** 2, 0., dt / 2 * subj_noise ** 2],
        #                                [0., motor_noise ** 2, 0.],
        #                                [dt / 2 * subj_noise ** 2, 0., subj_noise ** 2]]))

        Q = jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]])
        # Q = jnp.array([[1., -1., 0., 0.], [-1., 1., 0., 0.], [0., 0., 0., 0.], [0., 0., 0., 0.]])
        R = jnp.eye(B.shape[1]) * c

        # W = jnp.diag(jnp.array([sigma]))

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class SubjectiveVelocityDiffModel(System):
    def __init__(self, process_noise=1., c=0.5, motor_noise=0.5, subj_noise=1., sigma=6., dt=1. / 60):
        A = jnp.eye(2)
        B = jnp.array([[0.], [10. * dt]])
        C = jnp.array([[1., -1.]])

        V = jnp.diag(jnp.array([process_noise, motor_noise]))
        W = jnp.diag(jnp.array([sigma]))

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)

        A = jnp.array([[1., 0., dt], [0., 1., 0.], [0., 0., 1.]])
        B = jnp.array([[0.], [10. * dt], [0.]])
        C = jnp.array([[1., -1., 0.]])

        # V = jnp.diag(jnp.array([0., motor_noise, subj_noise]))
        V = linalg.cholesky(jnp.array([[dt ** 2 / 3 * subj_noise ** 2, 0., dt / 2 * subj_noise ** 2],
                                       [0., motor_noise ** 2, 0.],
                                       [dt / 2 * subj_noise ** 2, 0., subj_noise ** 2]]))

        Q = jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]])
        R = jnp.eye(B.shape[1]) * c

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class TemporalDelayModel(System):
    def __init__(self, system, delay=3):
        dyn = system.dynamics
        d = dyn.A.shape[0]

        A = linalg.block_diag(dyn.A, jnp.diag(jnp.zeros(d * delay))) + jnp.diag(jnp.ones(d * delay), k=-d)

        B = jnp.vstack([dyn.B] + [jnp.zeros_like(dyn.B)] * delay)

        C = jnp.hstack([jnp.zeros((dyn.C.shape[0], dyn.C.shape[1] * delay)), dyn.C])

        V = linalg.block_diag(dyn.V, jnp.diag(jnp.zeros(d * delay)))
        W = dyn.W

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)

        act = system.actor
        d = act.A.shape[0]

        A = linalg.block_diag(act.A, jnp.diag(jnp.zeros(d * delay))) + jnp.diag(jnp.ones(d * delay), k=-d)

        B = jnp.vstack([act.B] + [jnp.zeros_like(act.B)] * delay)

        C = jnp.hstack([jnp.zeros((act.C.shape[0], act.C.shape[1] * delay)), act.C])

        V = linalg.block_diag(act.V, jnp.diag(jnp.zeros(d * delay)))

        W = act.W

        Q = linalg.block_diag(act.Q, *[jnp.zeros_like(act.Q)] * delay)
        R = act.R
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DelayedSubjectiveVelocityModel(TemporalDelayModel):
    def __init__(self, process_noise=1., c=0.5, motor_noise=0.5, subj_noise=1., subj_vel_noise=10.,
                 sigma=6., prop_noise=3., dt=1. / 60):
        system = SubjectiveVelocityModel(process_noise=process_noise, c=c, motor_noise=motor_noise,
                                         subj_noise=subj_noise, subj_vel_noise=subj_vel_noise, sigma=sigma,
                                         prop_noise=prop_noise, dt=dt)

        super().__init__(system=system, delay=12)


if __name__ == '__main__':
    from jax import random
    import matplotlib.pyplot as plt

    from lqg.ccg import xcorr

    for model in [
        SubjectiveVelocityModel(process_noise=1., sigma=6., c=.1, motor_noise=.5, subj_noise=1.,
                                subj_vel_noise=10.),
        SubjectiveVelocityModel(process_noise=1., sigma=6., c=.1, motor_noise=.5, subj_noise=1.,
                                subj_vel_noise=0.)
    ]:  # ,
        # TemporalDelayModel(SubjectiveVelocityModel(process_noise=1., sigma=6., c=.5, motor_noise=.5, subj_noise=1.,
        #                                            subj_vel_noise=3.), delay=12)]:

        x = model.simulate(rng_key=random.PRNGKey(0), n=20, T=500)

        # plt.plot(x[:, 0, 0])
        # plt.plot(x[:, 0, 1])
        # plt.show()

        lags, correls = xcorr(jnp.diff(x[..., 1], axis=0).T, jnp.diff(x[..., 0], axis=0).T, maxlags=120)
        plt.plot(lags[100:], correls.mean(axis=0)[100:])

    plt.axhline(0, linestyle="--", color="gray")
    plt.show()

    # import numpy as np
    #
    # rw_std = 1.
    # dt = 1. / 60.
    # A = np.array([[1., dt], [0., 1.]])
    # V = np.diag(np.array([rw_std, 0.]))
    #
    # dyn = Dynamics(A, None, None, V)
    # x = dyn.simulate(n=20, T=1000)
    #
    # Vs = np.diag(np.array([0.8, .5]))
    #
    # dyns = Dynamics(A, None, None, Vs)
    # y = dyns.simulate(n=20, T=1000)
    #
    # f, ax = plt.subplots(2)
    # ax[0].plot(x[..., 0])
    # ax[1].plot(y[..., 0])
    # plt.show()

    # mcmc = infer(x, model=model, num_samples=1000, num_warmup=500)
    # import arviz as az
    #
    # az.plot_pair(mcmc.get_samples(), kind="hexbin")
    # plt.show()

    # import inspect
    #
    # signature = inspect.signature(DiffModel.__init__)



from lqg.tracking.basic import OneDimModel as BoundedActor, CostlessModel as OptimalActor
from lqg.tracking.kf import TrackingFilter as IdealObserver
from lqg.tracking.subjective import SubjectiveVelocityModel as SubjectiveActor
from lqg.tracking.eye import DampedSpringModel, DampedSpringVelocityModel, DampedSpringTwoDimModel, \
    DampedSpringDiffModel, DampedSpringSubjectiveVelocityModel, DampedSpringSubjectiveModel, \
    DampedSpringTwoDimSubjectiveModel, DampedSpringTwoDimFullModel, DampedSpringTrackingFilter, \
    DampedSpringCostlessModel, DampedSpringTwoDimCostlessModel



from dataclasses import dataclass
import jax.numpy as jnp
import numpyro.distributions as dist
from jax import vmap, random
from jax.lax import scan

from lqg.riccati import kalman_gain, control_law


@dataclass
class Dynamics:
    A: jnp.array
    B: jnp.array
    C: jnp.array
    V: jnp.array
    W: jnp.array


@dataclass
class Actor:
    A: jnp.array
    B: jnp.array
    C: jnp.array
    V: jnp.array
    W: jnp.array
    Q: jnp.array
    R: jnp.array

    def L(self, T):
        return control_law(self.A, self.B, self.Q, self.R, T=T)

    def K(self, T):
        return kalman_gain(self.A, self.C, self.V @ self.V.T, self.W @ self.W.T, T=T)


class System:
    def __init__(self, actor: Actor, dynamics: Dynamics):
        self.actor = actor
        self.dynamics = dynamics

    @property
    def xdim(self):
        """ State dimensionality

        Returns:
            int: dimensionality of state
        """
        return self.dynamics.A.shape[0]

    @property
    def ydim(self):
        """ Observation dimensionality

        Returns:
            int: dimensionality of observation
        """
        return self.dynamics.C.shape[0]

    @property
    def bdim(self):
        """ Belief dimensionality

        Returns:
            int: dimensionality of belief
        """
        return self.actor.A.shape[0]

    @property
    def udim(self):
        """ Action dimensionality

        Returns:
            int: dimensionality of action
        """
        return self.dynamics.B.shape[1]

    def simulate(self, rng_key, n=1, T=100, x0=None, xhat0=None, return_all=False):
        """ Simulate n trials

        Args:
            rng_key (jax.random.PRNGKey): random number generator key
            n (int): number of trials
            T (int): number of time steps
            x0 (jnp.array): initial state
            xhat0 (jnp.array): initial belief
            return_all (bool): return estimates, controls and observations as well

        Returns:
            jnp.array (T, n, d)
        """
        L = self.actor.L(T=T)
        K = self.actor.K(T=T)

        def simulate_trial(rng_key, T=100, x0=None, xhat0=None):
            """ Simulate a single trial

            Args:
                rng_key (jax.random.PRNGKey): random number generator key
                T (int): number of time steps
                x0 (jnp.array): initial state
                xhat0 (jnp.array): initial belief

            Returns:
                jnp.array, jnp.array, jnp.array, jnp.array: x (states), x_hat (estimates), y, u
            """

            x0 = jnp.zeros(self.xdim) if x0 is None else x0
            xhat0 = jnp.zeros(self.bdim) if xhat0 is None else xhat0

            # generate standard normal noise terms
            rng_key, subkey = random.split(rng_key)
            epsilon = random.normal(subkey, shape=(T, self.xdim))
            rng_key, subkey = random.split(rng_key)
            eta = random.normal(subkey, shape=(T, self.ydim))

            def loop(carry, t):
                x, x_hat = carry

                # compute control based on agent's current belief
                u = - L @ x_hat

                # apply dynamics
                x = self.dynamics.A @ x + self.dynamics.B @ u + self.dynamics.V @ epsilon[t]

                # generate observation
                y = self.dynamics.C @ x + self.dynamics.W @ eta[t]

                # update agent's belief
                x_pred = self.actor.A @ x_hat + self.actor.B @ u
                x_hat = x_pred + K @ (y - self.actor.C @ x_pred)

                return (x, x_hat), (x, x_hat, y, u)

            _, (x, x_hat, y, u) = scan(loop, (x0, xhat0), jnp.arange(1, T))

            return jnp.vstack([x0, x]), jnp.vstack([xhat0, x_hat]), \
                   jnp.vstack([self.dynamics.C @ x0 + self.dynamics.W @ eta[0]]), u

        # simulate n trials
        x, x_hat, y, u = vmap(lambda key: simulate_trial(key, T=T, x0=x0, xhat0=xhat0),
                              out_axes=1)(random.split(rng_key, num=n))

        if return_all:
            return x, x_hat, y, u
        else:
            return x

    def conditional_moments(self, x, mu0=None):
        """ Conditional distribution p(x | theta)

                Args:
                    self: LQG
                    x: time series of shape T (time steps), n (trials), d (dimensionality)

                Returns:
                    numpyro.distributions.MultivariateNormal
                """
        T, n, d = x.shape

        L = self.actor.L(T)

        K = self.actor.K(T)

        F = jnp.vstack(
            [jnp.hstack([self.dynamics.A,
                         -self.dynamics.B @ L]),
             jnp.hstack([K @ self.dynamics.C @ self.dynamics.A,
                         self.actor.A - self.actor.B @ L - K @ self.actor.C @ self.actor.A])])

        G = jnp.vstack([jnp.hstack([self.dynamics.V, jnp.zeros_like(self.dynamics.C.T)]),
                        jnp.hstack([K @ self.dynamics.C @ self.dynamics.V, K @ self.dynamics.W])])

        mu = jnp.zeros((n, self.dynamics.A.shape[0] + self.actor.A.shape[0])) if mu0 is None else mu0
        Sigma = G @ G.T

        def f(carry, xt):
            mu, Sigma = carry

            mu = mu @ F.T + (xt - mu[:, :d]) @ jnp.linalg.inv(Sigma[:d, :d]).T @ (F @ Sigma)[:, :d].T

            Sigma = F @ Sigma @ F.T + G @ G.T - (F @ Sigma)[:, :d] @ jnp.linalg.inv(Sigma[:d, :d]) @ (Sigma @ F.T)[:d,
                                                                                                     :]
            return (mu, Sigma), (mu, Sigma)

        _, (mu, Sigma) = scan(f, (mu, Sigma), x)
        return mu, Sigma

    def conditional_distribution(self, x):
        T, n, d = x.shape
        mu, Sigma = self.conditional_moments(x)
        return dist.MultivariateNormal(mu[:, :, :d].transpose((1, 0, 2)), Sigma[:, :d, :d])

    def log_likelihood(self, x):
        return self.conditional_distribution(x[:-1]).log_prob(x[1:].transpose((1, 0, 2)))


class LQG(System):
    def __init__(self, A, B, C, V, W, Q, R):
        dynamics = Dynamics(A, B, C, V, W)
        actor = Actor(A, B, C, V, W, Q, R)
        super().__init__(actor=actor, dynamics=dynamics)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    dt = 1. / 60.

    # parameters
    action_variability = 0.5
    sigma = 6.
    sigma_prop = 3.
    action_cost = 0.5

    A = jnp.eye(2)
    B = jnp.array([[0.], [dt]])
    V = jnp.diag(jnp.array([1., action_variability]))

    C = jnp.eye(2)
    W = jnp.diag(jnp.array([sigma, sigma_prop]))

    Q = jnp.array([[1., -1.],
                   [-1., 1]])

    R = jnp.eye(1) * action_cost

    lqg = System(actor=Actor(A, B, C, V, W, Q, R),
                 dynamics=Dynamics(A, B, C, V, W))

    x = lqg.simulate(random.PRNGKey(0), x0=jnp.zeros(2), n=10, T=1000)

    plt.plot(x[:, 0, 0])
    plt.plot(x[:, 0, 1])
    plt.show()



from jax import numpy as jnp

from lqg.kalman import KalmanFilter
from lqg.model import Dynamics


class TrackingFilter(KalmanFilter):
    def __init__(self, dim=1, process_noise=1., sigma=6., dt=1. / 60.):
        A = jnp.eye(dim)
        C = jnp.eye(dim)
        V = jnp.eye(dim) * process_noise
        W = jnp.eye(dim) * sigma

        super().__init__(Dynamics(A, None, C, V, W))


class TwoDimTrackingFilter(KalmanFilter):
    def __init__(self, process_noise=1., sigma_v=6., sigma_h=6., dt=1. / 60.):
        A = jnp.eye(2)
        C = jnp.eye(2)
        V = jnp.eye(2) * process_noise
        W = jnp.diag(jnp.array([sigma_h, sigma_v]))
        super().__init__(Dynamics(A, None, C, V, W))


if __name__ == '__main__':
    from jax import random
    import matplotlib.pyplot as plt
    import arviz as az

    from lqg.infer import infer

    kf = TrackingFilter()

    x = kf.simulate(random.PRNGKey(113), n=20, T=500)

    plt.plot(x[:, 0, 0])
    plt.plot(x[:, 0, 1])
    plt.show()

    mcmc = infer(x, model=TrackingFilter, num_samples=5_000, num_warmup=2_000)

    data = az.convert_to_inference_data(mcmc)
    az.plot_trace(data)
    plt.show()



import jax.numpy as jnp
from jax import random, lax
from numpyro import optim
from numpyro.infer import NUTS, MCMC, SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoBNAFNormal
from numpyro.infer.reparam import NeuTraReparam

from lqg.infer.models import lifted_model as lqg_model
from lqg.tracking import BoundedActor


def infer(x, num_samples, num_warmup, model=BoundedActor, numpyro_fn=lqg_model, process_noise=1., dt=1. / 60,
          method="nuts", progress_bar=True, num_chains=1, seed=0, **fixed):
    if method == "nuts":
        # setup kernel
        nuts_kernel = NUTS(numpyro_fn)

    elif method == "neutra":
        # learn normalizing flow
        guide = AutoBNAFNormal(numpyro_fn)
        svi = SVI(numpyro_fn, guide, optim.Adam(0.003), Trace_ELBO())
        svi_state = svi.init(random.PRNGKey(seed), x, model, process_noise, dt, **fixed)

        last_state, losses = lax.scan(lambda state, i: svi.update(state, x, model, process_noise, dt, **fixed),
                                      svi_state, jnp.zeros(10_000))

        # reparametrize model
        neutra = NeuTraReparam(guide, svi.get_params(last_state))
        nuts_kernel = NUTS(neutra.reparam(numpyro_fn))

    else:
        raise ValueError("Please specify a valid inference method (nuts, neutra).")

    # run NUTS
    mcmc = MCMC(nuts_kernel, num_samples=num_samples, num_warmup=num_warmup, progress_bar=progress_bar,
                num_chains=num_chains)
    mcmc.run(random.PRNGKey(seed), x, model, process_noise, dt, **fixed)

    return mcmc



import numpy as np
import scipy.io as spio


# methods for properly reading matlab structs
# taken from https://stackoverflow.com/a/8832212

def loadmat(filename):
    """
    this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    """
    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


def _check_keys(dict):
    """
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries
    """
    for key in dict:
        if isinstance(dict[key], spio.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict


def _todict(matobj):
    """
    A recursive function which constructs from matobjects nested dictionaries
    """
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, spio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict


def load_tracking_data(delay=12, clip=120, subtract_mean=True):
    """ Load tracking data from Bonnen et al. (2015)

    Args:
        delay: temporal delay between target and response
        clip: clip the first n time steps
        subtract_mean: subtract the mean of both trajectories?

    Returns:
        np.array: data from tracking task
    """

    # factor to convert between pixels and acrmin
    arcscale = 1.32

    # load matlab file
    mat = loadmat("../data/data.mat")

    # get target blob widths
    sigma = (mat["sigma"] * arcscale).round()
    sigmas = np.unique(sigma)

    # convert to numpy.float32
    target = mat["target"].astype(np.float32)
    mouse = mat["response"].astype(np.float32)

    # apply delay and clip the first couple of time steps
    if delay:
        target = target[:, clip:-delay]
        mouse = mouse[:, clip + delay:]
    else:
        target = target[:, clip:]
        mouse = mouse[:, clip:]

    # subtract the mean from both trajectories (in accordance with the analysis by Bonnen et al., 2015)
    # see https://github.com/kbonnen/BonnenEtAl2015_KalmanFilterCode/blob/master/tracking2graphs.m
    if subtract_mean:
        target = target - np.mean(target, axis=1, keepdims=True)
        mouse = mouse - np.mean(mouse, axis=1, keepdims=True)

    # stack data from all trials
    data = np.stack(
        [np.array(
            [target[np.where(sigma == blob_width)[0], :],
             mouse[np.where(sigma == blob_width)[0], :]])
            for blob_width in sigmas])

    # bring in right shape for our analysis methods
    data = data.transpose(0, 2, 3, 1)

    # subtract first time step in x dimension
    data = data - data[:, :, 0, 0][:, :, np.newaxis, np.newaxis]

    return data, sigmas



import inspect
import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist

from lqg.infer.prior import prior


def get_model_params(model_class):
    init_signature = inspect.signature(model_class.__init__)

    parameters = {}
    for name, param in init_signature.parameters.items():
        if name not in ["self", "dim", "dt", "process_noise", "delay", "covar"]:
            parameters[param.name] = param.default

    return parameters


def lqg_model(x, model_type, process_noise=1., dt=1. / 60, **fixed_params):
    params = {}
    for name, default in get_model_params(model_type).items():
        if name in fixed_params:
            params[name] = fixed_params[name]
        else:
            params[name] = numpyro.param(name, jnp.array(default), constraint=dist.constraints.positive)

    # dim = x.shape[2] // 2
    lqg = model_type(process_noise=process_noise, dt=dt, **params)

    if x is not None:
        # numpyro.sample("x0", dist.MultivariateNormal(jnp.zeros(lqg.A.shape[0]), V), obs=x[:, 0])
        numpyro.sample("x", lqg.conditional_distribution(x[:-1]),
                       obs=x[1:].transpose((1, 0, 2)))


def common_lqg_model(x, model_type, process_noise=1., dt=1. / 60., **fixed_params):
    Nc, T, N, d = x.shape

    params = {}
    for name, default in get_model_params(model_type).items():
        if not name == "sigma":
            if name in fixed_params:
                params[name] = fixed_params[name]
            else:
                params[name] = numpyro.param(name, jnp.array(default), constraint=dist.constraints.positive)

    for n in range(Nc):
        xn = x[n]
        # observation noise
        sigma_n = numpyro.param(f"sigma_{n}", jnp.array(1.), constraint=dist.constraints.positive)

        lqg = model_type(process_noise=process_noise, dt=dt, sigma=sigma_n, **params)

        numpyro.sample(f"x_{n}",
                       lqg.conditional_distribution(xn[:-1]),
                       obs=xn[1:].transpose((1, 0, 2)))


def loo_lqg_model(x, model_type, process_noise, dt=1. / 60., **fixed_params):
    Nc, T, N, d = x.shape

    params = {}
    for name, default in get_model_params(model_type).items():
        if name in fixed_params:
            params[name] = fixed_params[name]
        else:
            params[name] = numpyro.param(name, jnp.array(default), constraint=dist.constraints.positive)

    for n in range(Nc):
        xn = x[n]

        lqg = model_type(process_noise=process_noise[n], dt=dt, **params)

        numpyro.sample(f"x_{n}",
                       lqg.conditional_distribution(xn[:-1]),
                       obs=xn[1:].transpose((1, 0, 2)))


default_prior = prior()

# apply priors
lifted_model = numpyro.handlers.lift(lqg_model, prior=default_prior)
lifted_common_model = numpyro.handlers.lift(common_lqg_model, prior=default_prior)
lifted_loo_model = numpyro.handlers.lift(loo_lqg_model, prior=default_prior)


def correlated_noise_model(x, model_type, process_noise=1., dt=1. / 60, **fixed_params):
    d = x.shape[2] // 2

    params = {}
    for name, default in get_model_params(model_type).items():
        if name in fixed_params:
            params[name] = fixed_params[name]
        else:
            params[name] = numpyro.sample(name, default_prior[name])

    L = numpyro.sample("chol", dist.LKJCholesky(d))
    cor = numpyro.deterministic("cor", (L @ L.T)[1, 0])
    sigma = numpyro.sample("sigma", dist.HalfCauchy(jnp.ones(d) * 50.))

    lqg = model_type(process_noise=process_noise, dt=dt, covar=jnp.diag(sigma) @ L, **params)

    if x is not None:
        # numpyro.sample("x0", dist.MultivariateNormal(jnp.zeros(lqg.A.shape[0]), V), obs=x[:, 0])
        numpyro.sample("x", lqg.conditional_distribution(x[:-1]),
                       obs=x[1:].transpose((1, 0, 2)))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lqg",
    version="0.1.0",
    author="Dominik Straub",
    author_email="dominik.straub@tu-darmstadt.de",
    description="(Inverse) optimal control for linear quadratic Gaussian systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dominikstrb/lqg",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "numpy>=1.20.3",
        "scipy>=1.6.3",
        "matplotlib==3.4.2",
        "numpyro==0.9.2",
        "jax[cpu]==0.3.15",
        "arviz>=0.11.2",
    ],
)



from jax import numpy as jnp
from jax.scipy import linalg

from lqg.model import System, Dynamics, Actor


class DimModel(System):
    def __init__(self, dim=3, process_noise=1.0, motor_noise=0.5,
                 sigma=6.0, prop_noise=6.0, c=1.0, dt=1. / 60.):
        self.dim = dim
        self.process_noise = process_noise
        # dimensionality
        d = 2 * dim
        # dynamics model
        A = jnp.eye(d)
        B = dt * linalg.block_diag(*[jnp.array([[0.], [10.]])] * dim)

        # observation model
        C = jnp.eye(d)

        # noise model
        V = jnp.diag(jnp.array([process_noise, motor_noise] * dim))
        W = jnp.diag(jnp.array([sigma, prop_noise] * dim))

        # cost function
        Q = linalg.block_diag(*[jnp.array([[1., -1.], [-1., 1.]])] * dim)
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class OneDimModel(DimModel):
    def __init__(self, process_noise=1.0, motor_noise=0.5,
                 sigma=6.0, prop_noise=6.0, c=1.0, dt=1. / 60):
        super().__init__(dim=1, process_noise=process_noise,
                         motor_noise=motor_noise,
                         sigma=sigma, prop_noise=prop_noise, c=c, dt=dt)


class NoiseFreeModel(DimModel):
    def __init__(self, process_noise=1.0,
                 sigma=6.0, prop_noise=6.0, c=1.0, dt=1. / 60):
        super().__init__(dim=1, process_noise=process_noise,
                         motor_noise=1e-4,
                         sigma=sigma, prop_noise=prop_noise, c=c, dt=dt)


class CostlessModel(DimModel):
    def __init__(self, process_noise=1.0, motor_noise=0.5,
                 sigma=6.0, prop_noise=6.0, dt=1. / 60):
        super().__init__(dim=1, process_noise=process_noise,
                         motor_noise=motor_noise,
                         sigma=sigma, prop_noise=prop_noise, c=0, dt=dt)


class TwoDimModel(DimModel):
    def __init__(self, process_noise=1.0, motor_noise=0.5,
                 sigma=6.0, prop_noise=6.0, c=1.0, dt=1. / 60):
        super().__init__(dim=2, process_noise=process_noise,
                         motor_noise=motor_noise,
                         sigma=sigma, prop_noise=prop_noise, c=c, dt=dt)


class DiffModel(System):
    def __init__(self, dim=1, process_noise=1.0, motor_noise=0.5,
                 sigma=6.0, c=1.0, dt=1. / 60.):
        self.dim = dim
        self.process_noise = process_noise

        # dynamics model
        A = jnp.eye(2 * dim)
        B = dt * linalg.block_diag(*[jnp.array([[0.], [10.]])] * dim)

        # observation model
        C = linalg.block_diag(*[jnp.array([[1., -1.]])] * dim)

        # noise model
        V = jnp.diag(jnp.array([process_noise, motor_noise] * dim))
        W = jnp.diag(jnp.array([sigma] * dim))

        # cost function
        Q = linalg.block_diag(*[jnp.array([[1., -1.], [-1., 1.]])] * dim)
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class VelocityModel(System):
    def __init__(self, process_noise=1.0, c=1., motor_noise=0.5, prop_noise=3.0, sigma=6.0, dt=1. / 60.):
        A = linalg.block_diag(*[jnp.array([[1.]]), jnp.array([[1., dt], [0., 1.]])])
        B = jnp.array([[0.], [0.], [10. * dt]])

        C = jnp.array([[1., 0, 0.], [0., 1., 0.]])

        # V = linalg.block_diag(*[jnp.array([[process_noise]]),
        #                         jnp.array([[1 / 3 * dt ** 3, 1 / 2 * dt**2], [1 / 2 * dt**2, dt]]) * motor_noise])
        V = jnp.diag(jnp.array([process_noise, motor_noise, 0.]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))

        # cost function
        Q = jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., c]])
        R = jnp.eye(B.shape[1]) * 0.

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class VelocityDiffModel(System):
    def __init__(self, process_noise=1.0, c=1., motor_noise=0.5, vel_noise=1., sigma=6.0, dt=1. / 60.):
        A = linalg.block_diag(*[jnp.array([[1.]]), jnp.array([[1., dt], [0., 1.]])])
        B = jnp.array([[0.], [0.], [10. * dt]])

        C = jnp.array([[1., -1., 0.]])

        # V = linalg.block_diag(*[jnp.array([[process_noise]]),
        #                         jnp.array([[1 / 3 * dt ** 2, 1 / 2 * dt], [1 / 2 * dt, 1.]]) * motor_noise])
        V = jnp.diag(jnp.array([process_noise, motor_noise, vel_noise]))
        W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., c]])
        R = jnp.eye(B.shape[1]) * 0.

        dyn = Dynamics(A=A, B=B, C=C, V=V, W=W)
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)
